//
// The program reads a file of points in the plane (the number of
// points followed by pairs of coordinates) it then computes the
// Delaunay triagulation of the given points and transforms it into
// planar-map representation where one can interactively ask point
// location queries
//
// a sample input file del.txt is attached
//

#include <CGAL/basic.h>
#include <iostream>

// Planar Map (DCEL)

#include <CGAL/Cartesian.h>
#include <CGAL/Pm_default_dcel.h>
#include <CGAL/Planar_map_2.h>
#include <CGAL/Pm_segment_exact_traits.h>
#include <CGAL/Pm_naive_point_location.h>
#include <CGAL/IO/leda_window.h>
#include <CGAL/leda_rational.h>

// for Delaunay
#include <CGAL/squared_distance_2.h>   // to avoid a g++ problem
#include <CGAL/Point_2.h>
#include <CGAL/predicates_on_points_2.h>
#include <CGAL/Triangle_2.h>
#include <CGAL/Segment_2.h>

#include <CGAL/Triangulation_short_names_2.h>
#include <CGAL/Triangulation_euclidean_traits_2.h>
#include <CGAL/Triangulation_2.h>
#include <CGAL/Delaunay_triangulation_2.h>

typedef leda_rational NT;
typedef CGAL::Cartesian<NT> R;
typedef CGAL::Pm_segment_exact_traits<R>   Traits;

typedef CGAL::Pm_default_dcel<Traits> DCEL;
typedef CGAL::Planar_map_2<DCEL, Traits> PM;
typedef CGAL::Pm_naive_point_location<PM> NaivePL;
typedef Traits::X_curve X_curve;
typedef Traits::Point Point;
typedef Traits::Vector Vector;

// Delaunay
typedef CGAL::Triangulation_euclidean_traits_2<R> Gt;
typedef CGAL::Triangulation_vertex_base_2<Gt> Vb;
typedef CGAL::Triangulation_face_base_2<Gt>  Fb;
typedef CGAL::Triangulation_default_data_structure_2<Gt,Vb,Fb> Tds;
typedef CGAL::Triangulation_2<Gt,Tds>  Triangulation_;
typedef CGAL::Delaunay_triangulation_2<Gt,Tds>  Delaunay;

typedef Triangulation_::Face  Face_;
typedef Triangulation_::Vertex Vertex_;
typedef Vertex_::Vertex_handle Vertex_handle_;
typedef Face_::Face_handle Face_handle_;

typedef Triangulation_::Face_circulator  Face_circulator_;
typedef Triangulation_::Vertex_circulator  Vertex_circulator_;

typedef Triangulation_::Locate_type Locate_type_;

typedef Triangulation_::Face_iterator  Face_iterator_;
typedef Triangulation_::Vertex_iterator  Vertex_iterator_;
typedef Triangulation_::Edge_iterator  Edge_iterator_;
typedef Triangulation_::Line_face_circulator  Line_face_circulator_;
typedef Triangulation_::Edge_circulator  Edge_circulator_;

void draw_planar_map(CGAL::Window_stream &W, PM &pm)
{
	PM::Halfedge_handle he;
	for (he = pm.halfedges_begin(); he != pm.halfedges_end(); he++)
	{
		W << he->curve();
	}
}

void draw_planar_map_face(CGAL::Window_stream &W, PM &pm, PM::Face_handle face)
{
	PM::Halfedge_handle he;
	PM::Holes_iterator holes;
	PM::Ccb_halfedge_circulator ccb_circ, ccb_circ_begin;		
	
	if (face->does_outer_ccb_exist()) 
	{
		ccb_circ_begin = face->outer_ccb();
		ccb_circ = ccb_circ_begin;
		do
		{
			W << ccb_circ->curve();
			ccb_circ++;
		} while (ccb_circ != ccb_circ_begin);
	}
    for (holes = (PM::Holes_iterator)(face->holes_begin()); holes != (PM::Holes_iterator)(face->holes_end()); holes++)
    {
        ccb_circ_begin = *holes;
        ccb_circ = ccb_circ_begin;
        do
        {
			W << ccb_circ->curve();
            ccb_circ++;
        } while (ccb_circ != ccb_circ_begin);
    }
}

template<class Iterator>
void calc_bounding_window(Iterator begin, Iterator end, double &minx, double &miny, double &maxx, double &maxy)
{
	minx = miny = maxx = maxy = 0;

	if (begin != end)
	{
		maxx = minx = CGAL::to_double(begin->x());
		maxy = miny = CGAL::to_double(begin->y());
	}

	Iterator it;
	for (it = begin; it != end; it++)
	{
		if (CGAL::to_double(it->x()) < minx)
			minx = CGAL::to_double(it->x());
		if (CGAL::to_double(it->x()) > maxx)
			maxx = CGAL::to_double(it->x());
		if (CGAL::to_double(it->y()) < miny)
			miny = CGAL::to_double(it->y());
		if (CGAL::to_double(it->y()) > maxy)
			maxy = CGAL::to_double(it->y());
	}
}

void fit_in_window(int width, int height, double &minx, double &miny, double &maxx, double &maxy)
{
	double diffx = maxx - minx;
	double diffy = maxy - miny;
	double centerx = (minx + maxx) / 2.0;
	double a;
	a = (double)width / (double)height * diffy;
	if (a < diffx)
		a = diffx;
	minx = centerx - a/2;
	maxx = centerx + a/2;
}

int main(int argc, char *argv[])
{
    double x,y;
    int b;

    Delaunay::Finite_edges_iterator eit;
    Delaunay::Finite_vertices_iterator vit;
    Delaunay dt;
    CGAL::Segment_2<R> s;

	if (argc < 2)
	{
		std::cout << "parameters: <input filename>" << std::endl;
		return 0;
	}

	// read points from input file
	int i, count;
	std::list<Point> points;
	std::ifstream inp(argv[1]);
	inp >> count;
	for (i = 0; i < count; i++)
	{
		inp >> x >> y;
		points.push_back(Point(x, y));
	}


	// build Delaunay triangulation of the input points
	std::list<Point>::iterator it;
	for (it = points.begin(); it != points.end(); it++)
	{
		dt.insert(*it);
	}

	// build the planar map of the triangulation
	PM pm;
    for (eit = dt.finite_edges_begin(); eit != dt.finite_edges_end(); eit++) 
	{
		s = dt.segment(*eit);
		pm.insert(s);
    }

	// find proper boundaries for the window that contain all points
	double minx, miny, maxx, maxy;
	calc_bounding_window(points.begin(), points.end(), minx, miny, maxx, maxy);
	fit_in_window(600, 550, minx, miny, maxx, maxy);

    // initialize window
    CGAL::Window_stream W(600, 600);
    W.init(minx, maxx, miny);
    W.set_mode(leda_src_mode);
    W.set_node_width(3);
    W.set_line_width(3);
    
    W.button("Quit", 4);
    W.display();
    
	draw_planar_map(W, pm);

    while(1)
    {

        // handle window events
        b = W.read_mouse(x,y);

        // draw map and selected edge
        W.clear();

        W << CGAL::BLACK;
		draw_planar_map(W, pm);

		// left mouse button pressed
		if (b == -1)
		{
			PM::Locate_type lt;
			PM::Halfedge_handle he;
			// execute point location query
			he = pm.locate(Traits::Point(x, y), lt);
	        W << CGAL::RED;
			draw_planar_map_face(W, pm, he->face());
		}

        if (b == 4) break;
    }

    return 0;
}
